import sys
from time import sleep

from mizue.printer import Printer
from mizue.util import Utility
from mizue.util.stoppable_thread import StoppableThread
from .progress_renderer_args import ProgressBarRendererArgs, SpinnerRendererArgs, LabelRendererArgs, \
    InfoSeparatorRendererArgs, InfoTextRendererArgs, PercentageRendererArgs


class Progress:
    def __init__(self, start: int = 0, end: int = 100, value: int = 0):
        self._active = False
        self._end = end
        self._info_text = ""
        self._interval = 0.1
        self._spinner = [
            "▹▹▹▹▹",
            "▸▹▹▹▹",
            "▹▸▹▹▹",
            "▹▹▸▹▹",
            "▹▹▹▸▹",
            "▹▹▹▹▸"
        ]
        self._spinner_end_symbol = "⠿"
        self._spinner_index = 0
        self._start = start
        self._thread = None
        self._value = value
        self._width = 10

        self.info_separator = " | "
        self.info_separator_renderer = lambda args: self._info_separator_renderer(
            args.separator)  # InfoSeparatorRendererArgs
        self.info_text_renderer = lambda args: self._info_text_renderer(args.text)  # InfoTextRendererArgs
        self.label = ""
        self.label_renderer = lambda args: args.label  # LabelRendererArgs
        self.percentage_renderer = lambda args: "{:.2f}%".format(args.percentage)  # PercentageRendererArgs
        self.spinner_renderer = lambda args: args.spinner  # SpinnerRendererArgs
        self.progress_bar_renderer = lambda args: self._progress_bar_renderer()  # ProgressBarRendererArgs

    def set_end_value(self, end: int) -> None:
        """Update the maximum value of the progress bar"""
        self._end = end

    def set_info(self, info: str) -> None:
        """Set the info text to be displayed after the progress bar"""
        self._info_text = info

    def set_update_interval(self, interval: float) -> None:
        """Set the update interval of the progress bar"""
        self._interval = interval

    def start(self) -> None:
        """Start the progress bar"""
        Utility.hide_cursor()
        self._thread = StoppableThread(target=self._print, args=())
        self._active = True
        self._thread.start()

    def stop(self) -> None:
        """Stop the progress bar"""
        sleep(1)
        self._active = False
        self._spinner_index = 0
        self._thread.join()
        Utility.show_cursor()

    def terminate(self) -> None:
        """Terminate the progress bar.
            This method is generally used when the progress bar is needed to be stopped (e.g. on Ctrl+C)"""
        sleep(1)
        self._active = False
        self._spinner_index = 0
        self._thread.stop()
        self._thread.join()
        Utility.show_cursor()

    def update_value(self, value: int) -> None:
        """Update the value of the progress bar"""
        self._value = value

    def _get_bar_full_width(self) -> int:
        percentage = self._value * 100 / self._end
        bar_width = int(percentage * self._width / 100)
        return bar_width

    def _get_progress_text(self) -> str:
        percentage_value = self._value * 100 / self._end
        percentage = self.percentage_renderer(PercentageRendererArgs(percentage_value, self._value))
        spinner_symbol = self.spinner_renderer(SpinnerRendererArgs(
            spinner=self._spinner[self._spinner_index % len(self._spinner)],
            value=self._value,
            percentage=percentage_value
        ))
        bar_string = self._progress_bar_renderer()
        bar = self.progress_bar_renderer(ProgressBarRendererArgs(
            percentage=percentage_value,
            text=bar_string,
            value=self._value,
            width=self._get_bar_full_width()
        ))
        label = self.label_renderer(LabelRendererArgs(
            label=self.label,
            value=self._value,
            percentage=percentage_value
        ))
        separator = self.info_separator_renderer(InfoSeparatorRendererArgs(
            separator=self.info_separator,
            value=self._value,
            percentage=percentage_value
        ))
        info_text = self.info_text_renderer(InfoTextRendererArgs(
            text=self._info_text,
            value=self._value,
            percentage=percentage_value
        ))
        progress_text = str.format("{}{} {} {}{}{}", label, bar, spinner_symbol, percentage, separator,
                                   info_text)
        text_length = len(Printer.strip_colors(progress_text))
        if text_length > Utility.get_terminal_width():
            progress_text = str.format("{}{} {} {}{}", label, bar, spinner_symbol, percentage, '')
            if text_length > Utility.get_terminal_width():
                progress_text = str.format("{}{} {} {}{}", '', bar, spinner_symbol, percentage, '')
        return progress_text

    def _info_separator_renderer(self, info_text: str) -> str:
        return self.info_separator if len(info_text) > 0 else ""

    @staticmethod
    def _info_text_renderer(info_text: str) -> str:
        return info_text

    def _print(self) -> None:
        while self._active:
            progress_text = self._get_progress_text()
            self._spinner_index += 1
            sys.stdout.write(
                u"\u001b[K")  # Erase from cursor to end of line [http://matthieu.benoit.free.fr/68hc11/vt100.htm]
            sys.stdout.write(
                u"\u001b[1000D" + progress_text)  # Move terminal cursor 1000 characters left (go to start of line)
            sys.stdout.flush()
            sleep(self._interval)

    def _progress_bar_renderer(self):
        bar_start = "⟪"
        bar_end = "⟫"
        width = self._get_bar_full_width()
        bar = bar_start + "◆" * int(width) + " " * int((self._width - width)) + bar_end
        return bar



from signal import signal, Signals


class SignalHandler:
    def __init__(self, signal: Signals, callback: callable):
        self._callback = callback
        self._signal = signal
        self._signal_received = False

    def __call__(self, *args, **kwargs):
        self._signal_received = True
        self._signal_handler()

    def is_signal_received(self):
        return self._signal_received

    def _signal_callback(self):
        self._signal_received = True
        self._callback()

    def _signal_handler(self):
        signal(self._signal, self._signal_callback)



from dataclasses import dataclass


@dataclass(frozen=True)
class ProgressData:
    downloaded: int
    filename: str
    filepath: str
    filesize: int
    percent: int
    finished: bool
    url: str
    uuid: str



import os
import re
from math import ceil, floor
from typing import Callable

from wcwidth import wcswidth, wcwidth

from mizue.printer import Printer
from mizue.util import Utility
from .alignment import Alignment
from .border_character_codes import BorderCharacterCodes
from .border_style import BorderStyle
from .cell_renderer_args import CellRendererArgs
from .column import Column
from .column_settings import ColumnSettings
from .row_border_position import RowBorderPosition


class Grid:
    def __init__(self, columns: list[ColumnSettings], data: list[list[str]]):
        self.border_color = None
        self.border_style = BorderStyle.BASIC
        self.cell_renderer: Callable[[CellRendererArgs], str] = lambda args: Grid._get_default_cell_renderer(args)
        self.columns = []
        self.data = data
        self._prepare_columns(columns)

    def print(self) -> None:
        """Print the grid"""
        print(self._buffer())

    def _buffer(self) -> str:
        buffer = [self._create_row_border(RowBorderPosition.TOP), os.linesep]
        title_list = [column.title for column in self.columns]
        buffer.append(self._create_row(title_list, True))
        buffer.append(os.linesep)
        buffer.append(self._create_row_border(RowBorderPosition.MIDDLE))
        buffer.append(os.linesep)
        for row in self.data:
            buffer.append(self._create_row(row, False))
            buffer.append(os.linesep)
        buffer.append(self._create_row_border(RowBorderPosition.BOTTOM))
        return "".join(buffer)

    def _create_row(self, row: list[str], is_header_row: bool) -> str:
        border_style = self._get_border_style()
        row_buffer = []
        for index, cell_value in enumerate(row):
            cell = str(cell_value)
            column = self.columns[index]

            renderer = self._get_cell_renderer(column)
            if renderer is not None:
                rendered_cell = renderer(CellRendererArgs(cell=cell, index=index, is_header=is_header_row,
                                                          width=column.width))

                rendered_cell = self._format_cell_with_colors(rendered_cell, column.width)
            else:
                rendered_cell = self._format_cell_with_colors(cell, column.width)

            border = Printer.format_hex(border_style.VERTICAL, self.border_color) \
                if self.border_color else border_style.VERTICAL
            if index == 0:
                row_buffer.append(f"{border}")

            row_buffer.append(" ")
            row_buffer.append(self._get_left_cell_space(column, self._get_raw_cell_text_after_rendering(rendered_cell)))
            row_buffer.append(rendered_cell)
            row_buffer.append(
                self._get_right_cell_space(column, self._get_raw_cell_text_after_rendering(rendered_cell)))
            row_buffer.append(" ")
            row_buffer.append(border)
        return "".join(row_buffer)

    def _create_row_border(self, position):
        dash_list = []
        border_style = self._get_border_style()
        if position is RowBorderPosition.TOP:
            left = border_style.TOPLEFT
            middle = border_style.TOPMIDDLE
            right = border_style.TOPRIGHT
        elif position is RowBorderPosition.BOTTOM:
            left = border_style.BOTTOMLEFT
            middle = border_style.BOTTOMMIDDLE
            right = border_style.BOTTOMRIGHT
        else:
            left = border_style.LEFTMIDDLE
            middle = border_style.MIDDLEMIDDLE
            right = border_style.RIGHTMIDDLE
        dash_list.append(left)
        for index, max_length in enumerate(list(map(lambda column: column.width, self.columns))):
            dash_list.append("".join([border_style.HORIZONTAL] * (max_length + 2)))
            if index != len(self.columns) - 1:
                dash_list.append(middle)
        dash_list.append(right)
        return Printer.format_hex("".join(dash_list), self.border_color) if self.border_color else "".join(dash_list)

    def _find_max_cell_width(self, column: Column) -> int:
        max_width = len(column.title)
        for row in self.data:
            cell = str(row[column.index])
            length = 0
            for char in str(cell):
                if Grid._is_wide_char(char):
                    length += 2
                else:
                    length += 1
            length = length + 1 if Grid._has_variation_selector(cell) else length
            max_width = max(max_width, length)
        return max_width

    def _format_cell_with_colors(self, rendered_cell: str, column_width: int) -> str:
        color_parts = self._split_text_into_color_parts(rendered_cell)
        if len(color_parts) == 0:
            return self._format_long_cell(rendered_cell, column_width)
        else:
            processed_width = 0
            formatted_parts = []
            for color_part in color_parts:
                text = color_part[1]
                text_width = wcswidth(text)
                if processed_width + text_width <= column_width:
                    formatted_text = f"{color_part[0]}{text}{color_part[2]}"
                    formatted_parts.append(formatted_text)
                    processed_width += text_width
                    if processed_width == column_width:
                        break
                else:
                    visible_text = self._format_long_cell(text, column_width - processed_width)
                    formatted_text = f"{color_part[0]}{visible_text}{color_part[2]}"
                    formatted_parts.append(formatted_text)
                    break
            return "".join(formatted_parts)

    @staticmethod
    def _format_long_cell(cell: str, col_width: int) -> str:
        if col_width <= 3:
            if col_width == 1:
                return cell[0] if wcwidth(cell[0]) == 1 else "…"
            if col_width == 2:
                if len(cell) == 1:
                    return cell[0] if wcwidth(cell[0]) == 1 else "…"
                return cell[0] + cell[1] if wcwidth(cell[0]) == 1 and wcwidth(cell[1]) == 1 else "…"
            if col_width == 3:
                if wcwidth(cell[0]) == 2:
                    return cell[0] + "…"
                if wcwidth(cell[0]) == 1 and wcwidth(cell[1]) == 1:
                    return cell[0] + cell[1] + "…"
                if wcwidth(cell[0]) == 1 and wcwidth(cell[1]) == 2:
                    return "…"
                if wcwidth(cell[0]) == 1 and wcwidth(cell[1]) == 0:  # symbol + variation selector
                    return cell[0] + "…"

        text_width = 0
        text_length = 0
        for char in cell:
            text_length += 1
            if Grid._is_wide_char(char):
                text_width += 2
            else:
                text_width += 1
            if text_width == col_width - 3 or text_width == col_width - 2:
                break

        has_any_wide_char = any([True for char in cell if Grid._is_wide_char(char)])
        if not has_any_wide_char:
            if len(cell) <= col_width:
                return cell
            return cell[:col_width - 1] + "…"
        else:
            first_part = cell[:text_length]
            first_part_original = cell[:text_length]
            first_part_terminal_width = Grid._get_terminal_width_of_cell(first_part)
            full_width = Grid._get_terminal_width_of_cell(cell)
            while first_part_terminal_width > col_width - 1:
                first_part = first_part[:-1]
                first_part_terminal_width = Grid._get_terminal_width_of_cell(first_part)
            return first_part + "…" \
                if len(first_part) < len(first_part_original) or full_width > col_width \
                else first_part

    def _get_border_style(self):
        if self.border_style == BorderStyle.SINGLE:
            return BorderCharacterCodes.Single
        if self.border_style == BorderStyle.DOUBLE:
            return BorderCharacterCodes.Double
        if self.border_style == BorderStyle.BASIC:
            return BorderCharacterCodes.Basic
        if self.border_style == BorderStyle.EMPTY:
            return BorderCharacterCodes.Empty
        return BorderCharacterCodes.Basic

    def _get_cell_renderer(self, column: Column):
        if column.renderer:
            return column.renderer
        return self.cell_renderer

    @staticmethod
    def _get_default_cell_renderer(args: CellRendererArgs) -> str:
        if args.is_header:
            return Printer.format_hex(args.cell, '#FFCC75')
        return args.cell

    @staticmethod
    def _get_left_cell_space(column: Column, cell: str) -> str:
        cell_terminal_width = Grid._get_terminal_width_of_cell(cell)
        if column.alignment == Alignment.RIGHT:
            return "".join([" "] * (column.width - cell_terminal_width))
        elif column.alignment == Alignment.CENTER:
            return "".join([" "] * int(floor((column.width - cell_terminal_width) / 2)))
        return ""

    @staticmethod
    def _get_raw_cell_text_after_rendering(rendered_cell: str) -> str:
        # remove all the color codes and other formatting codes, also remove ansi escape codes
        return re.sub(r"\x1b\[[0-9;]*m", "", rendered_cell)

    @staticmethod
    def _get_right_cell_space(column: Column, cell: str) -> str:
        cell_terminal_width = Grid._get_terminal_width_of_cell(cell)
        if column.alignment == Alignment.LEFT:
            return "".join([" "] * (column.width - cell_terminal_width))
        elif column.alignment == Alignment.CENTER:
            return "".join([" "] * int(ceil((column.width - cell_terminal_width) / 2)))
        return ""

    @staticmethod
    def _get_terminal_width_of_cell(text):
        return sum([2 if wcswidth(char) == 2 else 1 for char in text])

    @staticmethod
    def _get_visible_text(text: str, max_width: int) -> str:
        current_width = 0
        for ch in text:
            if Grid._is_wide_char(ch):
                current_width += 2
            else:
                current_width += 1
            if current_width > max_width:
                return text[:text.index(ch)]
        return text

    @staticmethod
    def _has_variation_selector(text: str) -> bool:
        return any(Grid._is_variation_selector(c) for c in text)

    @staticmethod
    def _is_variation_selector(char: str) -> bool:
        return 0xFE00 <= ord(char) <= 0xFE0F

    @staticmethod
    def _is_wide_char(char: str) -> bool:
        return wcswidth(char) > 1

    def _prepare_columns(self, column_data: list[ColumnSettings]):
        columns: list[Column] = []
        for i, column_setting in enumerate(column_data):
            column = Column(settings=column_setting)
            column.index = i
            column.width = column_setting["width"] if "width" in column_setting else self._find_max_cell_width(column)
            columns.append(column)
        self.columns = columns
        self._resize_columns_to_fit()

    def _resize_columns_to_fit2(self):
        terminal_width = Utility.get_terminal_width()
        total_column_width = sum(column.width for column in self.columns)
        if total_column_width > terminal_width:
            for cx in range(0, len(self.columns)):
                self.columns[cx].width = int((terminal_width * self.columns[cx].width) / total_column_width) - 4

    def _resize_columns_to_fit(self):
        terminal_width = Utility.get_terminal_width()
        new_column_width = int(terminal_width / len(self.columns))
        long_columns = [column for column in self.columns if column.width >= new_column_width]

        remaining_width = 0
        for column in self.columns:
            if column.width > new_column_width:
                column.width = new_column_width
            else:
                remaining_width += (new_column_width - column.width - 4 * len(self.columns))

        padding = int(remaining_width / len(long_columns)) if len(long_columns) > 0 else 0
        for column in long_columns:
            column.width += padding

    @staticmethod
    def _split_text_into_color_parts(text: str) -> list[tuple[str, str, str]]:
        matcher = re.compile(r"(\x1b\[[0-9;]*m)(.*?)(\x1b\[00m)")
        groups = matcher.findall(text)
        return groups if groups else []



import ctypes
import shutil


class _CursorInfo(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]


class Utility:

    @staticmethod
    def get_terminal_size() -> tuple[int, int]:
        """Returns the size of the terminal."""
        return shutil.get_terminal_size()

    @staticmethod
    def get_terminal_width() -> int:
        """Returns the width of the terminal."""
        return Utility.get_terminal_size()[0]

    @staticmethod
    def get_terminal_height() -> int:
        """Returns the height of the terminal."""
        return Utility.get_terminal_size()[1]

    @staticmethod
    def hide_cursor() -> None:
        """Hides the cursor."""
        info = _CursorInfo()
        info.visible = False
        ctypes.windll.kernel32.SetConsoleCursorInfo(ctypes.windll.kernel32.GetStdHandle(-11), ctypes.byref(info))

    @staticmethod
    def is_elevated() -> bool:
        """Returns whether the current process is elevated."""
        return ctypes.windll.shell32.IsUserAnAdmin() != 0

    @staticmethod
    def is_caps_lock_on() -> bool:
        """Returns whether the caps lock key is on."""
        return ctypes.windll.user32.GetKeyState(0x14) != 0

    @staticmethod
    def show_cursor() -> None:
        """Shows the cursor."""
        info = _CursorInfo()
        info.visible = True
        ctypes.windll.kernel32.SetConsoleCursorInfo(ctypes.windll.kernel32.GetStdHandle(-11), ctypes.byref(info))
